__version__ = '0.1.0'
__all__ = [
    'Geyser'
]

import argparse
from io import FileIO
from typing import Text, Callable
from pathlib import Path
from sys import path as sys_path
from inspect import signature, Parameter

from _geysercpp import *


class Geyser:
    _core = Core()

    @classmethod
    def version(cls) -> Text:
        return __version__

    @classmethod
    def core_build(cls) -> Text:
        return cls._core.compiler

    @classmethod
    def _register_class(cls, name: Text, clazz: type, auto_compose: bool):
        name = clazz.__name__ if name is None else name
        if auto_compose:
            new_dict = {}
            new_dict.update(clazz.__dict__)
            new_dict['__init__'] = Composable.__dict__['__init__']
            clazz = type(name, (Composable,), new_dict)
        return cls._core.register_class(name, clazz)

    @classmethod
    def _build_executable(cls, name: Text, func: Callable):
        name = ''.join(map(
            lambda it: it.capitalize(),
            func.__name__.split('_'),
        )) if name is None else name
        clazz = type(name, (Composable,), {
            '__call__': func,
            '__module__': func.__module__,
            '__init__': Composable.__init__,
        })
        return clazz

    @classmethod
    def _parse_profile(cls, format, fp) -> dict:
        if format == 'json':
            import json
            return json.load(fp)
        elif format == 'yaml':
            from ruamel import yaml
            return yaml.load(fp, yaml.SafeLoader)
        elif format == 'toml':
            import toml
            return toml.load(fp)
        else:
            raise ValueError('Unsupported format of profile')

    @classmethod
    def composable(cls, clazz=None, name: Text = None, auto_compose: bool = True, **kwargs):
        if clazz is None:
            def wrapper(clz):
                cls._register_class(name, clz, auto_compose)
                return clz

            return wrapper
        else:
            new_clazz = cls._safe_init_class(clazz)
            cls._register_class(name, new_clazz, False)

    @classmethod
    def _safe_init_class(cls, clazz):
        def safe_init(self, *args, **kwargs):
            in_args = kwargs.pop('__args__') if '__args__' in kwargs else []
            params = signature(clazz).parameters
            args_slot = None

            missing = []
            undefined = []

            for key, param in params.items():
                if param.kind == Parameter.POSITIONAL_OR_KEYWORD or param.kind == Parameter.KEYWORD_ONLY:
                    if param.name not in kwargs and param.name is Parameter.empty:
                        missing.append(param)
                elif param.kind == Parameter.POSITIONAL_ONLY:
                    if args_slot is None:
                        args_slot = []
                    args_slot.append(param)
                elif param.kind == Parameter.VAR_POSITIONAL:
                    args_slot = param

            for key, value in kwargs.items():
                if key not in params.keys():
                    undefined.append(key)

            for key, value in params.items():
                if value.kind == Parameter.VAR_KEYWORD:
                    undefined.clear()

            fill_args, fill_kwargs = [], {}

            if isinstance(args_slot, list) and len(args_slot) != len(in_args):
                raise ValueError(f'{len(args_slot)} argument(s) are/is needed, input {len(in_args)} argument(s).')
            else:
                fill_args.extend(in_args)

            if len(missing) > 0:
                missing_names = list(map(
                    lambda param: param.name,
                    missing
                ))
                raise ValueError(f'Argument(s) named {", ".join(missing_names)} is/are missing.')
            if len(undefined) > 0:
                fill_kwargs.update(filter(
                    lambda it: it[0] not in undefined,
                    kwargs.items()
                ))
            else:
                fill_kwargs.update(kwargs.items())
            clazz.__init__(self, *fill_args, **fill_kwargs)

        new_clazz = type(clazz.__name__, (clazz,), {
            '__init__': safe_init,
            '__module__': clazz.__module__
        })
        return new_clazz

    @classmethod
    def executable(cls, function=None, name: Text = None, **kwargs):
        if function is None:
            def wrapper(func):
                clz = cls._build_executable(name, func)
                cls._register_class(clz.__name__, clz, False)
                return clz

            return wrapper
        else:
            clz = cls._build_executable(name, function)
            cls._register_class(clz.__name__, clz, False)

    @classmethod
    def access(cls, reference: Text) -> type:
        return cls._core[reference]

    @classmethod
    def entry(cls):
        sys_path.append(str(Path('.').absolute()))
        parser = argparse.ArgumentParser(
            'geyser', description='Geyser: compose & execute python objects.'
        )
        parser.add_argument(
            'profile',
            nargs=1,
            type=argparse.FileType('r', encoding='utf8'),
            help='compose this profile'
        )
        parser.add_argument('-v', '--version', action='version', version=f'{__version__}')
        ns = parser.parse_args()
        if len(ns.profile) > 0:
            profile_file: FileIO = ns.profile[0]
            format_name = profile_file.readline().strip('\r\n #/')
            profile = cls._parse_profile(format_name, profile_file)
            compose = profile['__compose__']
            for key, value in compose.items():
                cls._core.compose(key, compose)
            execute = profile['__execute__']
            cls._core.execute(execute)

    @classmethod
    def object_count(cls):
        return cls._core.object_count

    @classmethod
    def class_count(cls):
        return cls._core.class_count

    @classmethod
    def references(cls):
        return cls._core.references


class Composable:
    def __init__(self, *args, **kwargs):
        for key, item in kwargs.items():
            setattr(self, key, item)

    def __mod__(self, item) -> bool:
        return hasattr(self, item)

    def __bool__(self) -> bool:
        return hasattr(self, '__call__')



from tempfile import TemporaryDirectory
from os import makedirs, environ
from pathlib import Path
from logging.config import dictConfig
from typing import Text
from uuid import uuid4
from datetime import datetime

from geyser import Geyser


@Geyser.composable()
class PathManager:
    _tempdir = TemporaryDirectory(suffix='_folder', prefix='geyser_')
    _homedir = Path.home().absolute()
    _curdir = Path('.').absolute()

    @classmethod
    def _makedirs_join(cls, *args, root_dir) -> Path:
        new_path = root_dir.joinpath(*args)
        try:
            makedirs(new_path.parent)
        except FileExistsError:
            pass
        return new_path

    def temporary(self, *args) -> Path:
        root_dir = Path(self._tempdir.name).absolute()
        return self._makedirs_join(*args, root_dir=root_dir)

    def home(self, *args) -> Path:
        root_dir = self._homedir.joinpath('geyser')
        return self._makedirs_join(*args, root_dir=root_dir)

    def current(self, *args) -> Path:
        root_dir = self._curdir / 'geyser'
        return self._makedirs_join(*args, root_dir=root_dir)


@Geyser.composable(auto_compose=False)
class EnvManager:
    def __init__(self, *args, **kwargs):
        if 'logger' in kwargs:
            self.logger = kwargs['logger']
            kwargs.pop('logger')

        for key, item in kwargs.items():
            environ[key] = item

    def __contains__(self, item):
        return environ.__contains__(item)

    def __getitem__(self, item):
        self.logger.debug(f'Get "{item}" from environ.')
        return environ.__getitem__(item)

    def __setitem__(self, key, value):
        self.logger.debug(f'Set "{key}" in environ to "{value}".')
        return environ.__setitem__(key, value)

    def __delitem__(self, key):
        self.logger.debug(f'Remove "{key}" in environ.')
        return environ.__delitem__(key)


@Geyser.executable()
def runtime_info(ctx):
    import platform

    ctx.logger.info(f'Geyser v{Geyser.version()}')
    ctx.logger.info(f'Geyser Core build {Geyser.core_build()}')
    ctx.logger.info(f'Python build {platform.python_compiler()} {platform.python_build()[1]}')
    ctx.logger.info(f'Operating System {platform.platform()}')
    return {
        'platform': platform,
        'version': Geyser.version(),
    }


@Geyser.composable(auto_compose=False)
class LogManager:
    def __init__(self, **kwargs):
        dictConfig(kwargs)


@Geyser.composable(auto_compose=False)
class IdManager:
    def __init__(self, annotation: Text, **kwargs):
        self._uuid = uuid4().hex
        self._timestamp = datetime.now().strftime('%Y%m%d%H%M%%D')
        self._annotation = annotation

    def __call__(self, *args, **kwargs):
        if self._annotation is None:
            return f'{self._timestamp}_{self._uuid}'
        else:
            return f'{self._annotation}_{self._timestamp}_{self._uuid}'
